%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Función MATLAB con el algoritmo desarrollado para el tratamiento de una
% imagen. Mas en concreto, se encarga de detectar si hay una pieza en el
% conveyor, y en caso afirmativo, detectar su color, tamaño y posición.
% Datos de Entrada:
%   Imagen Capturada de la Simulación
% Operaciones Realizadas:
%   Deteccion de pieza, tamaño, color y posición
% Datos de Salida:
%   Color,Tamaño y Posición de la Pieza detectada
%
% Creado por:
%   - Rodrigo Sancho García
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [Color,Area,Posicion] = Algoritmo_VisionArtificial(Imagen)
ConfiguraEspacioVision = evalin('base','ConfiguraEspacioVision');
CoordenadasCorte=[187.5 127.5 30 45]; %Coordenadas de la zona de interes
% Método 1 de pasar la imagen a escala de grises
Img=rgb2gray(Imagen);
% Método 2 de pasar la imagen a escala de grises (con umbral)
umb=graythresh(Imagen); 
ImgGray=im2bw(Imagen,umb);

% Configuración Zona a Detectar si el Usuario ha seleccionado esta opción
if ConfiguraEspacioVision==1
    imcrop(ImgGray);
    CoordenadasCorte = evalin('base','CoordenadasCorte');
end

%Recorte Zona interes
ImgGrayCut=imcrop(ImgGray,CoordenadasCorte);

% Deteccion de Bordes de la imagen con método 'Canny'
ImgBordes=edge(ImgGrayCut,'Canny');
% Método 2: Rastrear los límites de la región en la imagen binaria
[B,L]=bwboundaries(ImgGrayCut,'noholes');

% Rellenar huecos de la imagen anterior en la que se detectaron los bordes
ImgRellenada=imfill(ImgBordes,'holes');

% Tratamiento de la imagen, para marcar el cubo detectado y nombrarle
f=figure('visible','off');
S6=subplot(1,2,1);
set(S6,'Visible','off');

% Encuentra componentes conexas y las asigna etiquetas (1, 2, 3…)
% Devuelve una estructura con diversos campos, entre ellos:
%      cc.NumObjects:  número de objetos encontrados
CC = bwconncomp(ImgRellenada);

% L:  imagen de etiquetas  (0 fondo; 1, 2… las distintas comp. conexas)
% P.ej.: para separar objetos y rotarlos
L=labelmatrix(CC);

% Muestra imagen de etiquetas
% ('jet, 'k' para fondo negro)
RellenoColores=label2rgb(L, 'jet', 'k');
imshow(RellenoColores);

% Medir las propiedades de las regiones de la imagen procesada
propied=regionprops(ImgRellenada,'basic');
hold on

X=0;
Y=0;
AreaFigura=0;

% Bucle para detectar todos los objetos rectangulares de la imagen y
% marcarlos. Si detecta el cubo lo marca en verde y le asigna un nombre,
% los demás objetos serán marcados en rojo.
% Además, se obtiene el tamaño y posición (Centroide) del cubo detectado
for n=1:size(propied)
    rectangle('Position',propied(n).BoundingBox,'EdgeColor','r','LineWidth',2);
    centroide=propied(n).Centroid;
    Area=propied(n).Area;
    plot(centroide(1),centroide(2),'ko');
    if(centroide(1)<25 & centroide(1)>10 & centroide(2)<30 & centroide(2)>15 & Area<110)
        rectangle('Position',propied(n).BoundingBox,'EdgeColor','g','LineWidth',5);
        text(centroide(1)-4,centroide(2)+7,'Cubo','Color','g','Fontsize',15,'Fontweight', 'bold');
        AreaFigura=propied(n).Area;
        X=centroide(1);
        Y=centroide(2);
    end
end

% Asignación de las imagenes obtenidas en el Workspace, para posteriormente
% mostrarlas en la Interfaz de App Designer
assignin('base','Imagen',Imagen);
assignin('base','ImagenGrises',ImgGray);
assignin('base','ImagenRecortada',ImgGrayCut);
assignin('base','ImagenBordes',ImgBordes);
assignin('base','ImagenRellenada',ImgRellenada);

F = getframe(S6);
assignin('base','FiguraDetectada',F.cdata);

% Tratamiento de los Colores de la Imagen
ZonaInteres=imcrop(Imagen,CoordenadasCorte);
assignin('base','ZonaInteres',ZonaInteres);
ZonaInteresGris=rgb2gray(ZonaInteres);

% Divide las 3 capas de una imagen RGB
imR=double(ZonaInteres(:,:,1));
imG=double(ZonaInteres(:,:,2));
imB=double(ZonaInteres(:,:,3));
% Se eliminan el resto de colores de la imagen excepto Red ó Green ó Blue
% Creando unas máscaras para esos colores
% Fondo: Negro y Color correspondiente: Blanco
imagenR=imR-imG-imB;
imagenG=imG-imR-imB;
imagenB=imB-imR-imG;
% Asignación de las anteriores imagenes
assignin('base','imagenR',imagenR);
assignin('base','imagenG',imagenG);
assignin('base','imagenB',imagenB);

imagen_binaria=imagenR>50;
imagen_binaria1=imagenG>50;
imagen_binaria2=imagenB>50;

% Se realiza el filtrado medio de la imagen en dos dimensiones
img_filt_bin=medfilt2(imagen_binaria);
img_filt_bin1=medfilt2(imagen_binaria1);
img_filt_bin2=medfilt2(imagen_binaria2);

% Se filtran las imagenes, para mantener unicamente el color buscado
% (R,G,B)
mascara=1-img_filt_bin;
mascara1=1-img_filt_bin1;
mascara2=1-img_filt_bin2;

imagen_roja=double(ZonaInteresGris)/255;
imagen_verde=double(ZonaInteresGris) .* mascara/255;
imagen_azul=double(ZonaInteresGris) .* mascara/255;
img_final=cat(3,imagen_roja,imagen_verde,imagen_azul);

imagen_roja=double(ZonaInteresGris) .* mascara1/255;
imagen_verde=double(ZonaInteresGris)/255;
imagen_azul=double(ZonaInteresGris) .* mascara1/255;
img_final1=cat(3,imagen_roja,imagen_verde,imagen_azul);

imagen_roja=double(ZonaInteresGris) .* mascara2/255;
imagen_verde=double(ZonaInteresGris) .* mascara2/255;
imagen_azul=double(ZonaInteresGris)/255;
img_final2=cat(3,imagen_roja,imagen_verde,imagen_azul);

assignin('base','AislaRojo',img_final);
assignin('base','AislaVerde',img_final1);
assignin('base','AislaAzul',img_final2);

ColorFigura='NULL';

% Algoritmo Detección Color
% Se recorre todos los pixeles de la imagen (Para las 3 capas, RGB) y
% detecta si existen pixeles de los colores buscados
[Filas,Columnas] = size(ZonaInteres);
for i=1:Filas
    for j=1:(Columnas/3)
        VectorColores=impixel(ZonaInteres,j,i);
        %%%%%ROJO%%%%%
        if VectorColores(1)>100 & VectorColores(2)<40 & VectorColores(3)<40
            ColorFigura='Rojo';
        %%%%%VERDE%%%%%
        elseif VectorColores(1)<40 & VectorColores(2)>100 & VectorColores(3)<40
            ColorFigura='Verde';
        %%%%%AZUL%%%%%
        elseif VectorColores(1)<40 & VectorColores(2)<40 & VectorColores(3)>100
            ColorFigura='Azul';
        %%%%%AMARILLO%%%%%
        elseif VectorColores(1)>100 & VectorColores(2)>100 & VectorColores(3)<40
            ColorFigura='Amarillo';
        %%%%%MAGENTA%%%%%
        elseif VectorColores(1)>100 & VectorColores(2)<40 & VectorColores(3)>100
            ColorFigura='Magenta';
        %%%%%CYAN%%%%%
        elseif VectorColores(1)<40 & VectorColores(2)>100 & VectorColores(3)>100
            ColorFigura='Cyan';
        %%%%%NEGRO%%%%%
        elseif VectorColores(1)<40 & VectorColores(2)<40 & VectorColores(3)<30
            ColorFigura='Negro';
        end
    end
end

% Asignación del Tamaño, Posición y Color detectado
Area=AreaFigura;
Posicion=X;
Color=ColorFigura;

fprintf('El area de la figura es: %d\r\n',AreaFigura);
fprintf('La posicion de la figura es: "X=%d":"Y=%d"\r\n',X,Y);
fprintf('El color de la figura es: %s\r\n',ColorFigura);

end
