%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Función MATLAB para crear la estación del robot IRB120 (TFG_Rodrigo
% Sancho) y tras desplazar la pieza por el conveyor hasta el lugar deseado,
% se toma una imagen de la situación de la pieza, para posteriormente
% procesarla con visión artificial.
% Datos de Entrada:
%   Color,Tamaño y Posición Inicial del Cubo que hay que crear, para
%   realizar la simulación
% Operaciones Realizadas:
%   Creacion de la estación, simulacion del movimiento del cubo, captura de
%   dicho movimiento, y captura final de imagen para un posterior
%   procesamiento con visión artificial
% Datos de Salida:
%   Color,Tamaño y Posición Final del Cubo (Parámetros detectados mediante
%   visión artificial)
%
% Creado por:
%   - Rodrigo Sancho García
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Color,Area,Posicion]=EstacionMATLAB(Color,Size,PosicionX)
% Vacia lo datos anteriores del robot creado
clear robot rob
deg= pi/180; % Conversion Grados a Radianes

% Variables que determinan el movimiento del Cubo (Traslacion y Rotacion)
% Linspace: Generar un vector espaciado linealmente
% Zeros: Crear un array de ceros
nPts= 1e3;
t= linspace(0,10,nPts)';
x= zeros(nPts,1);
y= linspace(0,1,nPts)';
z= zeros(nPts,1);
rot=zeros(nPts,1);
Ini=zeros(nPts,1);

% Instrucción para modificar la herramienta o los objetos de trabajo.
copiar= @(orig, dest) eval(['! copy .\Visual\',orig,'.stl  ','.\Visual\',dest,'.stl']);

% Posición inicial del robot y del rastreador
q0= zeros(10,1);
pose= randn(3,3)*1e-3;

% Asignacion del Color y la Posición del Cubo elegidos por el Usuario en 
% la interfaz,a las variables utilizadas en la simulación
Pos=[0,1,0];
ColorCubo=Color;
PosicionFigura=PosicionX-1760; % Se resta una constante para ajustar a los valores de la simulación
Pieza1.base= [[PosicionFigura,1000,220]*1e-3, [0,0,0]*deg];

% Posición del robot
robot1= [[0,0,0]*1e-3, [0,0,0]*deg];

% Posición de las herramienta 1 respecto de la herramienta 0
copiar('Ventosa', 'tool1');
tool1.tcp= [[0,0,60]*1e-3, [0,0,0]*deg];
% Posición de las herramienta 2 respecto de la herramienta 1
copiar('Null', 'load1');
% Asignacion del tamaño elegido por el Usuario en la interfaz, a la
% simulacion del cubo. En función del tamaño de este, las posiciones 
% tomadas por el robot para la recogida y entrega del cubo en los conveyors
% son diferentes (Siempre busca el centro, pero su altura es diferente)
if Size==[0.2,0.2,0.2]
    copiar('CuboGrande','Figura');
    P_Recogida= [[PosicionFigura+650,0,420]*1e-3,[0,0,180]*deg];
    P_Entrega= [[-350,435,280]*1e-3,[0,0,180]*deg];
elseif Size==[0.15,0.15,0.15]
    copiar('CuboMediano','Figura');
    P_Recogida= [[PosicionFigura+650,0,370]*1e-3,[0,0,180]*deg];
    P_Entrega= [[-350,435,230]*1e-3,[0,0,180]*deg];
elseif Size==[0.1,0.1,0.1]
    copiar('CuboPequeno','Figura');
    P_Recogida= [[PosicionFigura+650,0,320]*1e-3,[0,0,180]*deg];
    P_Entrega= [[-350,435,180]*1e-3,[0,0,180]*deg];
end
load1.tcp= [[0,0,0]*1e-3, [0,0,0]*deg];

% Definición de los objetos de trabajo introducidos en la Estación
copiar('Conveyor_4', 'wobj1');
wobj1.base= [[1350,-400,-470]*1e-3, [90,0,0]*deg];
copiar('Cristal1_2', 'wobj2');
wobj2.base= [[2229,1612.5,804.60]*1e-3, [90,90,0]*deg];
copiar('Luz', 'wobj3');
wobj3.base= [[440, 900, 655]*1e-3, [0, 0, 180]*deg];
copiar('Luz_1', 'wobj4');
wobj4.base= [[595,900,655]*1e-3, [0,0,180]*deg];
copiar('Luz_2', 'wobj5');
wobj5.base= [[1120,900,655]*1e-3, [0,0,180]*deg];
copiar('Luz_2_1', 'wobj6');
wobj6.base= [[1275,900,655]*1e-3, [0,0,180]*deg];
copiar('Null', 'wobj7');
wobj7.base= [[-350,435,80]*1e-3, [0,0,0]*deg];
wobj8.base= [[-25,-25,0]*1e-3, [0,0,0]*deg];
copiar('Conveyor', 'wobj9');
wobj9.base= [[-1000,350,0]*1e-3, [0,0,0]*deg];
copiar('Conveyor', 'wobj10');
wobj10.base= [[-1000,-525,0]*1e-3, [0,0,0]*deg];
copiar('MesaTrabajo', 'wobj11');
wobj11.base= [[-1200,0,0]*1e-3, [0,0,0]*deg];
copiar('BloqueTrabajo', 'wobj12');
wobj12.base= [[-900,-50,0]*1e-3, [0,0,0]*deg];

% Asignación de todas las variables utilizadas durante la simulación
% en Workspace, para poder ser asignadas en el modelo de Simulink
assignin('base','ColorCubo',ColorCubo);
assignin('base','Pos',Pos);
assignin('base','q0',q0);
assignin('base','pose',pose);
assignin('base','tool1',tool1.tcp);
assignin('base','load1',load1.tcp);
assignin('base','Pieza1',Pieza1.base);
assignin('base','robot1',robot1);
assignin('base','wobj1',wobj1.base);
assignin('base','wobj2',wobj2.base);
assignin('base','wobj3',wobj3.base);
assignin('base','wobj4',wobj4.base);
assignin('base','wobj5',wobj5.base);
assignin('base','wobj6',wobj6.base);
assignin('base','wobj7',wobj7.base);
assignin('base','wobj8',wobj8.base);
assignin('base','wobj9',wobj9.base);
assignin('base','wobj10',wobj10.base);
assignin('base','wobj11',wobj11.base);
assignin('base','wobj12',wobj12.base);
assignin('base','P_Recogida',P_Recogida);
assignin('base','P_Entrega',P_Entrega);
assignin('base','PosicionFigura',PosicionFigura);

CoordenadaX=[t,x];CoordenadaY=[t,y];CoordenadaZ=[t,z];RotacionZ=[t,rot];Inicial=[t,Ini];
assignin('base','CoordenadaX',CoordenadaX);assignin('base','CoordenadaY',CoordenadaY);
assignin('base','CoordenadaZ',CoordenadaZ);assignin('base','RotacionZ',RotacionZ);
assignin('base','Inicial',Inicial);

% Ejecución Simulación del modelo de la estacion creado en Simulink
sim('Estacion_TFG_RodrigoSancho');
% Crea un video de la simulación
% NOTA IMPORTANTE %
% En Mechanical Explorer se debe tener la distribución de la vista en 4
% ventanas, para ello establezca: View > Layout > Four Standard Views
% El video se graba de la segunda vista, que es la vista aérea
smwritevideo('Estacion_TFG_RodrigoSancho','video','tile',2);
pause(15); % Pausa en lo que ejecuta el video grabado
v= VideoReader('video.avi'); % Lectura del video grabado
p= read(v, Inf); % Captura de imagen del último fotograma del video grabado

% Tratamiento de la Imagen capturada con visión artificial.
% Se obtiene el Color, Tamaño y Posición del Cubo, en el que caso de que
% exista y se detecte tal figura.
[Color,Area,Posicion] = Algoritmo_VisionArtificial(p);

end