classdef Hmat < handle
% Clase para transformacin de la matriz homognea
% Autor: Alberto Herreros
% Basado en las funciones de la librera Robotic Toolbox (Corke)
% Versin 7-Abril-2020
    
    properties
        T;
        deg;
    end % properties
    
    methods
        function this= Hmat(H)
        % Constructor:
        % Input: H matriz homognea/objeto Hmat/R rotacin/t traslacin
        
            if nargin==0
                H= [eye(3), zeros(3,1); zeros(1,3), 1];
            end
            
            if isa(H,'Hmat')
            % Hmat es handles. Se copia en un nuevo objeto 
                this= Hmat;
                this.T= H.T;
                this.deg= H.deg;
            else
                [n,m]= size(H);
                if n==4 && m==4
                   this.T= H;
                elseif n*m==9
                   this= Hmat;
                   this.R(H);
                elseif n*m==3
                   this= Hmat;
                   this.t(H);
                else
                   error('Mal argumento entrada');
                end
                this.deg= 1;
            end
        end

        function T= H(this, Pose)
        % Con argumentos: Introduce una T 
        % Sin argumentos: Muestra la T del objeto
            if nargin==1
                T= this.T;
            else
                this.T= Pose;
                T= Pose;
            end
        end
                
        function Deg(this)
        % Deg trabaja en grados
            this.deg= 1;
        end
        
        function Rad(this)
        % Trabaja en radianes
            this.deg= 0;
        end
        
        function disp(this)
        % Representa la matriz homognea
           Rt= this.T';
           fprintf('\n      R                | t\n'); 
           fprintf('-------------------------------\n');
           fprintf('%6.3f, %6.3f, %6.3f | %6.3f\n', Rt(1:12));
        end
        
        function Tinv= Inv(this)
        % Devuelve la T invertida     
            Tinv= inv(this.T);
            this.H(Tinv);
        end
        
        function obj3= mtimes(obj1, obj2)
        % Multiplica dos objetos. 
        % Devuelve un objeto
            obj1= Hmat(obj1);
            obj2= Hmat(obj2);
            obj3= Hmat;
            obj3.T= obj1.T*obj2.T;
        end
       
        function out= Plot(this, mag, name)
        % Muestra la posicin de los ejes en matlab
           if nargin==1
              mag= [1,2]; name=[];
           elseif nargin==2
              name= [];
           end
           if length(mag)==1
               mag= [mag,2];
           end
           eje= [0,0,0; mag(1),0,0; 0,mag(1),0; 0,0,mag(1)];
           [eje(:,1), eje(:,2),eje(:,3)]= this.Txyz(eje(:,1), eje(:,2),eje(:,3));
           color= {'r','g','b'};
           hold on;
           for i=1:3
              idx= [1,i+1];
              plot3(eje(idx,1), eje(idx,2), eje(idx,3), ...
                  color{i}, 'LineWidth', mag(2)); 
           end
           if ~isempty(name)
               text(eje(1,1), eje(1,2), eje(1,3), name);
           end
           view(3);
           if nargout~=0
              out= eje; 
           end
        end
        
        function T= T3pts(this, pts)
            % Genera un eje a partir de tres puntos
            %    * Dos puntos en el eje x
            %    * Un punto en el eje y
            
            if nargin==1
                error('T= T3pts(this, pts)');
            end
            
            [n,m]= size(pts);
            if n~=3 || m~= 3
                error('pts 3x3');
            end
            
            % Distancia de la recta del eje x con el punto del eje y
            dist= @(nu) sum(((1-nu)*pts(1,:)+ nu*pts(2,:)- pts(3,:)).^2);
            
            % Parmetro que minimiza esa distancia
            nu0= fminsearch(dist,rand);
            
            % Punto origen del eje
            p0= (1-nu0)*pts(1,:)+ nu0*pts(2,:);
            % Hallar los ejes normal (x) orientacin (y) aproximacin (z)
            n= (pts(2,:)- p0)'; n= n/norm(n);
            o= (pts(3,:)- p0)'; o= o/norm(o);
            a= cross(n,o);
            T= eye(4);
            T(1:3,1:3)= [n,o,a];
            T(1:3,4)= p0';
            this.T= T;
        end
        
        function [x, y, z]= Txyz(this, x, y, z) 
        % Trasforma vectores de puntos por T
        % La entrada puede ser en tres bloques x, y, z
        %     x, y, z pueden ser matriz o vector
        % La entrada puede ser en un bloque compacto xyz

            if nargin==4
                % x y z pueden tener forma vectorial o matricial
                [nx,mx]= size(x);
                [ny,my]= size(y);
                [nz,mz]= size(z);
                if (nx+ny+nz)/3~=nx || (mx+my+mz)/3~=mx
                    error('Bad dimension');
                end
                xyz= [x(:), y(:), z(:), ones(nx*mx,1)];
                xyz= (this.T*xyz')';
                x= reshape(xyz(:,1), [nx,mx]);
                y= reshape(xyz(:,2), [nx,mx]);
                z= reshape(xyz(:,3), [nx,mx]);
            elseif nargin==2
                % En x est el valor de xyz
                [n,m]= size(x);
                if m~= 3
                   error('Bad dimension'); 
                end
                xyz= [x, ones(n,1)];
                xyz= (this.T*xyz')';
                x= xyz(:,1:3);
            else
                error('[xt, yt, zt]= Txyz(this, x, y, z)');
            end
        end
        
        function vect= Tpose(this, vect, mag, form)
        % Trasforma vectores de puntos y ngulos o cuaternios por T
        % Entrada: vect la matriz de puntos y ngulos 
        %          mag: Magnitud para el plot
        %          form: es el formato de entrada (defecto tRzyx).  
        % Salida: vect trasformado. Puede dibujar los target en el espacio  
            if nargin<2
                error('vect= TtRzyx(this, vect, [mag], [form])')
            elseif nargin==2
                mag=[]; form= 'tRzyx';
            elseif nargin==3
                form= 'tRzyx';
            end
            n= size(vect,1);
            H= this.H;
            if strcmp(form,'tRzyx')
                for i= 1:n
                    this.H(H* this.tRzyx(vect(i,:)));
                    vect(i,:)= this.tRzyx;
                    if ~isempty(mag)
                        this.Plot(mag);
                    end
                end
            elseif strcmp(format,'tRxyz')
                for i= 1:n
                   this.H(H* this.tRxyz(vect(i,:)));
                   vect(i,:)= this.tRxyz;
                   if ~isempty(mag)
                       this.Plot(mag);
                   end
                end
            elseif strcmp(format,'tRzyz')
                for i= 1:n
                   this.H(H* this.tRzyz(vect(i,:)));
                   vect(i,:)= this.tRzyz;
                   if ~isempty(mag)
                       this.Plot(mag);
                   end
                end
            elseif strcmp(format,'tRyxz')
                for i= 1:n
                   this.H(H* this.tRyxz(vect(i,:)));
                   vect(i,:)= this.tRyxz;
                   if ~isempty(mag)
                       this.Plot(mag);
                   end
                end                
            elseif strcmp(format,'tQ')
                for i= 1:n
                   this.H(H* this.tQ(vect(i,:)));
                   vect(i,:)= this.tQ;
                   if ~isempty(mag)
                       this.Plot(mag);
                   end
                end
            else
                error('Mal formato'); 
            end
                                            
            this.H(H);     
        end

        
        
        function T= Offs(this, vec)
           % Modifica T en la posicin relativa de vec
           T= this.T;
           T(1:3, 4)= this.T(1:3, 4)+ vec(:);
        end
        
        
        function r = t(this, x, y, z)
        % Con argumentos: Convierte [x,y,z] en T
        % Sin argumentos: Muestra [x,y,z] de T
            if nargin==1
                r= this.T(1:3,4)';
                return
            elseif nargin==2
                if length(x)==3
                   y= x(2);
                   z= x(3);
                   x= x(1);
                else
                   error('r = t([x, y, z])');   
                end
            elseif nargin~=4
                error('r = t(x, y, z)');
            end
            r =[1, 0, 0, x;
                0, 1, 0, y;
                0, 0, 1, z;
                0, 0, 0, 1];
            this.T= r;
        end
        
        function r= R(this, Rot)
        % Con argumentos: Introduce una rotacin R en T
        % Sin argumentos: Muestra la matriz de rotacin de T
           if nargin==1
              r= this.T(1:3,1:3);
           else
              this.T(1:3,1:3)= Rot;
           end
        end
        
        function r = Rx(this, t)
        % Introduce una rotacin x en T
            if this.deg
                t= t*pi/180;
            end
            ct = cos(t);
            st = sin(t);
            r = [1	0	0   0
                0	ct	-st 0
                0	st	ct  0
                0   0   0   1];
            this.T= r;
        end
        
        function r = Ry(this, t)
        % Introduce una rotacin y en T
            if this.deg
                t= t*pi/180;
            end
            ct = cos(t);
            st = sin(t);
            r = [ct	0	st  0
                0	1	0   0
                -st	0	ct  0
                0   0   0   1];
            this.T= r;
        end
        
        function r = Rz(this, t)
        % Introduce una rotacin z en T
            if this.deg
                t= t*pi/180;
            end
            ct = cos(t);
            st = sin(t);
            r = [ct	-st	0    0
                st	ct	0    0
                0	0	1    0
                0   0   0    1];
            this.T= r;
        end
        
        function T= DH(this, DHm, joints, type)
        % Convierte una matriz DH (DHm) en T
        %   joints: Cinemtica directa 
        %   type: Tipo de articulacin 
        %         (rotacional (defecto) o prismtica)
   
            if nargin==1
               error('T= DH(this, DHm, [joints, type])');
            elseif nargin==2
                [n,~]= size(DHm);
                joints= zeros(1,n);
                type= [];
            elseif nargin==3
                type= [];
            end
            
            [n,m]= size(DHm);           
            % Definir ejes rotacionales y prismticos
            qR= zeros(1,n); qP= zeros(1,n);
            if isempty(type)
               qR= joints;
            else
               for i= 1:n
                  if type(i)=='R'
                      qR(i)= joints(i);
                  else 
                      qP(i)= joints(i);
                  end
               end                                
            end
            
            if m~=4
                error('Bad dimension DH matrix');
            end
            this.T= eye(4);
            for i= 1:n
               this.T= this.T* this.Rz(DHm(i,1)+qR(i))*this.t(0,0,DHm(i,2)+ qP(i))* ...
                       this.t(DHm(i,3),0,0)*this.Rx(DHm(i,4));
            end
            T= this.T;
        end
        
        function PlotDH(this, DHm, joints, mag, type)
            % Dos argumentos: Muestra en pantalla el DH
            % Tres argumentos: Cinemtica directa (joints)
            % mag: magnitud de los  ejes del robot
            % type: Tipo de articulacin 'R' rotacional 'P' prismtica 
            if nargin==1
                error('PlotDH(this, DHm, [joints, mag, type])');
            elseif nargin==2
                [n,~]= size(DHm);
                joints= zeros(1,n);
                mag= 0.1;
                type= [];
            elseif nargin==3
                mag= 0.1;
                type= [];
            elseif nargin==4
                type= [];
            end
            
            [n,m]= size(DHm);
            if m~=4
                error('Bad dimension DH matrix');
            end
            
            figure(1); clf; hold on;
            [az, el]= view; view(az,el);
           % axis([-0.2, 0.5, -0.5, 0.5, 0, 0.8])
            col= {'b','g','r','c','m','y'};
            T= Hmat();
            T.Plot(mag, 'T0');
            pEje0= T.t;
            
            for i= 1:size(DHm,1)
                T.H(T.DH(DHm(1:i,:), joints(1:i), type));
                pEje= T.t;
                plot3([pEje0(1), pEje(1)], [pEje0(2), pEje(2)], [pEje0(3), pEje(3)], ...
                    'Color',  col{rem(i,6)+1} ,'LineWidth', 5 )
                pEje0= pEje;
                T.Plot(mag, sprintf('T%d', i));
            end
            this.T= T.T; 
        end
        
        
        function [DHopt, err]= IDH(this, w, Nt)
        % [DHopt, err]= IDH(this, w, Nt)
        % Obtiene los parmetros DH que optimizan la matriz homognea
        %     w: peso de cada componente [x,y,z,Rx,Ry,Rz]
        %     Nt: Repeticiones del algoritmo optimizacin local 
            if nargin==1
               w= [ones(1,3)*1e3, ones(1,3)];
               Nt= 10;
            elseif nargin==2
               Nt= 10;
            end
                        
            F= @(x) IDHerr(x, this.H, w);
            DHopt= [0, 0, 0, 0];
            for i= 1:Nt
                DHopt= fminsearch(F, DHopt);
            end 
            err= F(DHopt);
        end
        
        
        function out = tRzyz(this, vect)
        % Con argumentos: Transforma [x,y,z,Rz,Ry,Rz] en T
        %                 Transforma [Rz,Ry,Rz] en T
        % Sin argumentos: Transforma T en [x,y,z,Rz,Ry,Rz]
        % Peter Corke
            if nargin==1
                R = this.T(1:3,1:3);
                if abs(R(1,3)) < eps && abs(R(2,3)) < eps
                    % singularity
                    euler(1) = 0;
                    sp = 0;
                    cp = 1;
                    euler(2) = atan2(cp*R(1,3) + sp*R(2,3), R(3,3));
                    euler(3) = atan2(-sp * R(1,1) + cp * R(2,1), -sp*R(1,2) + cp*R(2,2));
                else
                    euler(1) = atan2(R(2,3), R(1,3));
                    sp = sin(euler(1));
                    cp = cos(euler(1));
                    euler(2) = atan2(cp*R(1,3) + sp*R(2,3), R(3,3));
                    euler(3) = atan2(-sp * R(1,1) + cp * R(2,1), -sp*R(1,2) + cp*R(2,2));
                end
                
                if this.deg
                    euler = euler * 180/pi;
                end
                xyzRzyz= [this.T(1:3,4)', euler];
                if nargout==0
                    if this.deg==1
                       fprintf('[%.3f,%.3f,%.3f],[%.0f,%.0f,%.0f] deg\n', xyzRzyz);
                    else
                       fprintf('[%.3f,%.3f,%.3f],[%.3f,%.3f,%.3f] rad\n', xyzRzyz);
                    end
                else
                    out= xyzRzyz;
                end
            else
                if length(vect)==3
                   vect= [0,0,0,vect];
                elseif length(vect)~=6
                   error('Vector with 6 component');
                end                
                this.T= this.t(vect(1),vect(2),vect(3))* ...
                    this.Rz(vect(4))*this.Ry(vect(5))* this.Rz(vect(6));
                out= this.T;
            end
        end
        
        function out= tRxyz(this, vect)
        % Con argumentos: Transforma [x,y,z,Rx,Ry,Rz] en T
        %                 Transforma [Rx,Ry,Rz] en T
        % Sin argumentos: Transforma T en [x,y,z,Rx,Ry,Rz]
        % Peter Corke
            if nargin==1
                R= this.T(1:3,1:3);
                if abs(abs(R(1,3)) - 1) < eps  % when |R13| == 1
                    % singularity
                    rpy(1) = 0;  % roll is zero
                    if R(1,3) > 0
                        rpy(3) = atan2( R(3,2), R(2,2));   % R+Y
                    else
                        rpy(3) = -atan2( R(2,1), R(3,1));   % R-Y
                    end
                    rpy(2) = asin(R(1,3));
                else
                    rpy(1) = -atan2(R(1,2), R(1,1));
                    rpy(3) = -atan2(R(2,3), R(3,3));
                    
                    rpy(2) = atan2(R(1,3)*cos(rpy(1)), R(1,1));
                end
                
                if this.deg
                    rpy = rpy * 180/pi;
                end
                
                xyzRxyz= [this.T(1:3,4)',rpy(3:-1:1)];
                
                if nargout==0
                    if this.deg==1
                       fprintf('[%.3f,%.3f,%.3f],[%.0f,%.0f,%.0f] deg\n', xyzRxyz);
                    else
                       fprintf('[%.3f,%.3f,%.3f],[%.3f,%.3f,%.3f] rad\n', xyzRxyz);
                    end
                else
                    out= xyzRxyz;
                end
            else
                if length(vect)==3
                   vect= [0,0,0,vect];
                elseif length(vect)~=6
                   error('Vector with 6 component');
                end                                
                this.T= this.t(vect(1),vect(2),vect(3))* ...
                    this.Rx(vect(4))*this.Ry(vect(5))* this.Rz(vect(6));
                out= this.T;
            end
        end
        
        function out= tRzyx(this, vect)
        % Con argumentos: Transforma [x,y,z,Rz,Ry,Rx] en T
        %                 Transforma [Rz,Ry,Rx] en T
        % Sin argumentos: Transforma T en [x,y,z,Rz,Ry,Rx]
        % Peter Corke
            
            if nargin==1
% Corke
%                 R= this.T(1:3,1:3);
%                 
%                 if abs(abs(R(3,1)) - 1) < eps  % when |R31| == 1
%                     % singularity
%                     
%                     rpy(1) = 0;     % roll is zero
%                     if R(3,1) < 0
%                         rpy(3) = -atan2(R(1,2), R(1,3));  % R-Y
%                     else
%                         rpy(3) = atan2(-R(1,2), -R(1,3));  % R+Y
%                     end
%                     rpy(2) = -asin(R(3,1));
%                 else
%                     rpy(1) = atan2(R(3,2), R(3,3));  % R
%                     rpy(3) = atan2(R(2,1), R(1,1));  % Y
%                     
%                     rpy(2) = atan2(-R(3,1)*cos(rpy(1)), R(3,3));
%                 end
%                 
%                 if this.deg
%                     rpy = rpy * 180/pi;
%                 end
%                 
%                 xyzRzyx= [this.T(1:3,4)',rpy(3:-1:1)];

                eul= tform2eul(this.T);
                if this.deg
                   eul= eul* 180/pi;
                end
                xyzRzyx= [this.T(1:3,4)',eul];
                if nargout==0
                    if this.deg==1
                       fprintf('[%.3f,%.3f,%.3f],[%.0f,%.0f,%.0f] deg\n', xyzRzyx);
                    else
                       fprintf('[%.3f,%.3f,%.3f],[%.3f,%.3f,%.3f] rad\n', xyzRzyx);
                    end
                else
                    out= xyzRzyx;
                end
            else
                if length(vect)==3
                   vect= [0,0,0,vect];
                elseif length(vect)~=6
                   error('Vector with 6 component');
                end                
                
                this.T= this.t(vect(1),vect(2),vect(3))* ...
                    this.Rz(vect(4))*this.Ry(vect(5))* this.Rx(vect(6));
                out= this.T;
            end
        end
        
        
        function out= tRyxz(this, vect)
        % Con argumentos: Transforma [x,y,z,Ry,Rx,Rz] en T
        %                 Transforma [Ry,Rx,Rz] en T
        % Sin argumentos: Transforma T en [x,y,z,Ry,Rx,Rz]
        % Peter Corke
            if nargin==1
                %  YXZ for camara
                
                R= this.T(1:3,1:3);
                
                if abs(abs(R(2,3)) - 1) < eps  % when |R23| == 1
                    % singularity
                    rpy(1) = 0;
                    if R(2,3) < 0
                        rpy(3) = -atan2(R(3,1), R(1,1));   % R-Y
                    else
                        rpy(3) = atan2(-R(3,1), -R(3,2));   % R+Y
                    end
                    
                    rpy(2) = -asin(R(2,3));    % P
                else
                    rpy(1) = atan2(R(2,1), R(2,2));
                    rpy(3) = atan2(R(1,3), R(3,3));
                    
                    rpy(2) = atan2(-cos(rpy(1))*R(2,3), R(2,2));
                end
                
                if this.deg
                    rpy = rpy * 180/pi;
                end
                
                xyzRyxz= [this.T(1:3,4)',rpy(3:-1:1)];
                if nargout==0
                    if this.deg==1
                       fprintf('[%.3f,%.3f,%.3f],[%.0f,%.0f,%.0f] deg\n', xyzRyxz);
                    else
                       fprintf('[%.3f,%.3f,%.3f],[%.3f,%.3f,%.3f] rad\n', xyzRyxz);
                    end
                else
                    out= xyzRyxz;
                end
            else
                if length(vect)==3
                   vect= [0,0,0,vect];
                elseif length(vect)~=6
                   error('Vector with 3/6 component');
                end                
                
                this.T= this.t(vect(1),vect(2),vect(3))* ...
                    this.Ry(vect(4))*this.Rx(vect(5))* this.Rz(vect(6));
                out= this.T;
            end
        end
        
        function out= tQ(this, Q)
        % Con argumentos: Transforma [x,y,z,q1,q2,q3,q4] en T
        %                 Transforma [q1,q2,q3,q4] en T
        % Sin argumentos: Transforma T en [x,y,z,q1,q2,q3,q4]
        % Peter Corke
            
            if nargin==1
% Corke
%                 R= this.T(1:3,1:3);
%                 
%                 % Calculate all elements of symmetric K matrix
%                 K11 = R(1,1) - R(2,2) - R(3,3);
%                 K12 = R(1,2) + R(2,1);
%                 K13 = R(1,3) + R(3,1);
%                 K14 = R(3,2) - R(2,3);
%                 
%                 K22 = R(2,2) - R(1,1) - R(3,3);
%                 K23 = R(2,3) + R(3,2);
%                 K24 = R(1,3) - R(3,1);
%                 
%                 K33 = R(3,3) - R(1,1) - R(2,2);
%                 K34 = R(2,1) - R(1,2);
%                 
%                 K44 = R(1,1) + R(2,2) + R(3,3);
%                 
%                 % Construct K matrix according to paper
%                 K = [...
%                     K11,    K12,    K13,    K14;
%                     K12,    K22,    K23,    K24;
%                     K13,    K23,    K33,    K34;
%                     K14,    K24,    K34,    K44];
%                 
%                 K = K ./ 3;
%                 
%                 % and eigenvectors. The eigenvector corresponding to the largest eigenvalue
%                 % is the unit quaternion representing the same rotation.
% 
%                 [eigVec,eigVal] = eig(K(:,:),'vector');
%                 [~,maxIdx] = max(real(eigVal));
%                 quat = real([eigVec(4,maxIdx) eigVec(1,maxIdx) eigVec(2,maxIdx) eigVec(3,maxIdx)]);
%                 
%                 % By convention, always keep scalar quaternion element positive.
%                 % Note that this does not change the rotation that is represented
%                 % by the unit quaternion, since q and -q denote the same rotation.
%                 if quat(1) < 0
%                     quat = -quat;
%                 end
                quat= tform2quat(this.T);
                out= [this.T(1:3,4)', quat];
                if nargout==0
                    fprintf('[[%.3f,%.3f,%.3f],[%.2f,%.2f,%.2f,%.2f]]', ...
                        out(1), out(2), out(3), out(4), out(5), out(6), out(7));
                end
            else
                if length(Q)==4
                   Q= [0,0,0,Q];
                elseif length(Q)~=7
                   error('Vector with 7 component');
                end                
                                
                P= Q(1:3);
                Q= Q(4:7);
                Q = Q/norm(Q); % Ensure Q has unit norm
                
                w = Q(1); x = Q(2); y = Q(3); z = Q(4);
                w2 = w^2; x2 = x^2; y2 = y^2; z2 = z^2;
                xy = x*y; xz = x*z; yz = y*z;
                wx = w*x; wy = w*y; wz = w*z;
                
                this.T = [w2+x2-y2-z2 , 2*(xy - wz) , 2*(wy + xz) ,  P(1)
                    2*(wz + xy) , w2-x2+y2-z2 , 2*(yz - wx) ,  P(2)
                    2*(xz - wy) , 2*(wx + yz) , w2-x2-y2+z2 ,  P(3)
                    0      ,       0     ,       0     ,  1];
                
                out= this.T;
            end
        end
        
    end % methods
    
end % class


%% Funcin de costo para IDH

function err= IDHerr(x, T, w)
   T= Hmat(T); vT= T.tRxyz;
   T.DH(x); vDH= T.tRxyz; 
   err= sum(((vDH- vT).*w).^2);
end