%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Clase para manejar el robot IRB120 desde MATLAB
%
% Creado por mi tutor de TFG:
%   - Alberto Herreros albher@eis.uva.es
% Modificadas ciertas partes, para el caso utilizado en este TFG por:
%   - Rodrigo Sancho García
% Versión: 10-3-2021
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

classdef Kin < handle
    
    properties
       robot
       q
       tool
       h
       ejes
       psim
       save
    end %properties
    
    methods
        function this= Kin(robot, mdl, tool)
            if nargin==0
                error('this= Kin(robot, [mdl])');
            elseif nargin==1
                mdl= [];
            end
            robot.DataFormat= 'column';
            this.robot= robot;
            this.h= Hmat(); this.h.Rad();
            % Parámetros por defecto
            if nargin==3
                this.tool= tool;
            else
                this.tool= robot.BodyNames{end};
            end
            %fprintf('Herramienta: %s\n', this.tool);
            %this.q= [0,0,0,0,0,0];
            this.q= [0,0,0,0,0,0,0,0,0,0];
            this.ejes.axis= [-700,700,-700,700,-100,1000]*1e-3;
            this.ejes.view= [120,30];
            if isempty(mdl)
                this.psim.mdl= [];
            else
               this.Sim(mdl, 0.3); 
               %fprintf('Simulación. Vel= %.3f m/s\n', this.psim.value);
            end
            this.save.q= []; % Salva joint
            this.save.t= []; % Salva tiempo
            this.save.pose= []; % Salva pose
            this.save.value= 0; % 0 no graba, 1 graba
            this.Show();
        end
        
        function Body(this, name, body0, Htool, stl, Hstl, type)
           if nargin<6
               error('Tool(this, name, body0, Htool, stlname, Hstl, [type])')
           elseif nargin==6
               type= 'fixed';
           end
           body1= robotics.RigidBody(name);
           % Se crea un eje fijo
           jtool1 = robotics.Joint(name,type);
           % setFixedTransform permite matrices homogéneas
           setFixedTransform(jtool1, Htool);
           body1.Joint= jtool1;
           % Se une al cuerpo final del robot 'tool0'
           addBody(this.robot, body1, body0);
           % Definir la posición de lapiz con respecto al eje del tool1
           if ~isempty(Hstl)
               addVisual(this.robot.Bodies{end}, "Mesh", stl, Hstl);
           end
           Show(this);
        end
        
        
        function DelBody(this,name)
        % Borra un body del multi-body
            removeBody(this.robot, name);
            if strcmp(this.tool,name)==1
                this.tool= this.robot.BodyNames{end};
            end
            Show(this);
        end
                
        function Tool(this, name, body0, Htool, stl, Hstl)
        % Introduce una herramienta en el multi-body
            if nargin==2
                this.tool= name;
                return
            end
            if nargin<4
                error('Tool(this, name, body0, Htool[, stl, Hstl])') 
            elseif nargin==4
                stl= ''; Hstl= [];
            end
            this.Body(name, body0, Htool, stl, Hstl);
            this.tool= name;
        end
        
        function Wobj(this, name, Hwobj, stl, Hstl)
            if nargin<3
               error('Wobj(this, name, Hwobj[, stl, Hstl])')
            elseif nargin==4
                stl='';
                Hstl= [];
            end
            this.Body(name, this.robot.BaseName, Hwobj, stl, Hstl);
        end
        
        
        function robot= Robot(this)
            robot= this.robot;
        end
        
        
        function pose= Pose(this, q)
            if nargin==1
               q= this.q; 
            end
            nq= size(q, 1);
            pose= zeros(nq,6);
            for i= 1:nq
               this.h.H(getTransform(this.robot, q(i,:)', this.tool));
               pose(i,:)= this.h.tRzyx;
            end
        end
        
        function q= Joint(this, pose, w)
           if nargin==1
              q= this.q;
              return
           elseif nargin==2
               w= ones(1,6);
           end
           ik = robotics.InverseKinematics('RigidBodyTree', this.robot);
           npose= size(pose, 1);
           q= zeros(npose, length(this.q));
           q0= this.q';
           %pose(:,4:6)= rotm2eul(eul2rotm(pose(:,4:6)));
           for i= 1:npose
              H= this.h.tRzyx(pose(i,:));
              q0= ik(this.tool, H, w, q0);
              q(i,:)= q0';
           end
        end
        
        function q= MoveAbsJ(this, qN, nPts)
            if nargin<2
                error('q= MoveAbsJ(this, q1, [nPts])')
            elseif nargin==2
                nPts= 2;
            end
            if nPts>2
                nq= length(this.q);
                q= zeros(nPts, nq);
                % Interpolación de q
                for i= 1:nq
                    q(:,i)= linspace(this.q(1,i), qN(1,i), nPts)';
                end
            else
                q= [this.q; qN];
            end
            this.Show(q);
            this.q= q(end,:);
        end
        
        function q= MoveJ(this, poseN, Hwobj, nPts)
            if nargin<2
                error('q= MoveJ(this, poseN, [Hwobj], [nPts])')
            elseif nargin==2
                Hwobj=[]; nPts= 2;
            elseif nargin==3
                nPts= 2;
            end

            % Se desea no cambiar el giro actual
            [fil, col]= size(poseN);
            if col==3
               pose0= this.Pose(this.q);
               poseN(:, 4:6)= repmat(pose0(1,4:6), [fil,1]);
               w= [0,0,0,1,1,1];
            else
               w= ones(1,6);
            end
                       
            if ~isempty(Hwobj)
               this.h.H(Hwobj);
               poseN= this.h.Tpose(poseN);
            end
            
            % Solver con la estructura actual
            nq= length(this.q);
            if nPts>2
                qN= this.Joint(poseN(1,:),w);
                q= zeros(nPts, nq);
                % Interpolación de q
                for i= 1:nq
                    q(:,i)= linspace(this.q(1,i), qN(1,i), nPts)';
                end
            else
                q= [this.q; this.Joint(poseN,w)];
            end
            this.Show(q, poseN);
            this.q= q(end,:);
        end
        
        function q= MoveL(this, pose1, Hwobj, nPts)
            if nargin<2
                error('q= MoveL(this, pose1, [Hwobj], [nPts])');
            elseif nargin==2
                Hwobj= []; nPts= 10;
            elseif nargin==3
                nPts= 10;
            end
            
            pose0= this.Pose(this.q);
            if length(pose1)==3
                pose1(4:6)= pose0(4:6);
                w= [0,0,0,1,1,1];
            else
                w= ones(1,6);
            end
            
            if ~isempty(Hwobj)
               this.h.H(Hwobj);
               pose1= this.h.Tpose(pose1);
            end           
            
            poseN= zeros(nPts,6); 
            % Interpolación en todos los elementos
            for i= 1:6
               poseN(:,i)= linspace(pose0(i), pose1(i), nPts)';
            end
            q= this.Joint(poseN,w);
            this.Show(q, poseN);
            this.q= q(end, :);
        end
        
        function MoveC(this, y1, x1, Hwobj, nPts)
            
            if nargin<3 
                error('q= MoveC(this, x2, y1, [Hwobj], [nPts])');
            elseif nargin==3
                Hwobj= []; nPts= 20;
            elseif nargin==4
                nPts= 20;
            end
            
            if ~isempty(Hwobj)
               this.h.H(Hwobj);
               [x1(1),x1(2),x1(3)]= this.h.Txyz(x1(1),x1(2),x1(3));
               [y1(1),y1(2),y1(3)]= this.h.Txyz(y1(1),y1(2),y1(3));
            end
            
            pose1= this.Pose;
            x2= pose1(1:3);
            % Tres puntos
            pts= [x1; x2; y1];
            H1= this.h.T3pts(pts);
            %eje= this.h.tRzyx;
            pts= [pts, repmat(pose1(:,4:6), [3,1])];
            % Mover los puntos al nuevo eje 
            % x1, x2 solo tienen componente x
            % y1 solo tiene componente y
            this.h.Inv;
            p3pts= this.h.Tpose(pts);
            % Se mueve este eje de forma que x1 y x2 sean simétricos y la distancia de
            % los tres puntos al origen sea la misma
            x1= p3pts(1,1); x2= p3pts(2,1); y1= p3pts(3,2);
            xc= -(x1+x2)/2;
            yc= (x1^2- y1^2+ 2*x1*xc)/2/y1;
            H2= this.h.t([-xc, -yc, 0]);
            this.h.H(H1*H2);
            this.h.Inv;
            % Puntos iniciales en el nuevo eje x1 y x2 simétricos y todos a la misma
            % distancia del origen
            Tpts= this.h.Tpose(pts);
            % Radio del arco es la distancia a los puntos 
            % la misma distancia a todos los tres puntos
            radio= norm(Tpts(1,1:3));
            % Ángulo para trazar el arco desde el x1 al x2 pasando por y1
            theta0= asin(yc/radio);
            % Puntos del arco en el nuevo eje y en el eje origen
            pose= [zeros(nPts,3), repmat(Tpts(1,4:6),[nPts,1])];
            theta= linspace(theta0, pi-theta0, nPts);
            for i= 1:nPts
               pose(i,1)= radio*cos(theta(i));
               pose(i,2)= radio*sin(theta(i));
            end
            % Mover por Joint de los puntos generados
            this.MoveJ(pose(:,1:6), H1*H2);
        end
        
        function Show(this, q, pose)
            if nargin==1
                q= []; pose= [];
            elseif nargin==2
                pose=[];
            end
            if isempty(q)
                q= this.q;
            end
            if isempty(this.psim.mdl) % Gráfica matlab
               for i= 1:size(q,1)
                  hold off
                  show(this.robot, q(i,:)');
                  axis(this.ejes.axis);
                  if ~isempty(pose)
                     hold on
                     this.h.H(eye(4));
                     this.h.Tpose(pose,0.05);
                  end
                  view(this.ejes.view(1),this.ejes.view(2));
                  drawnow
               end
           else % Simulación Simulink multi-body
               if this.psim.type=='T'
                   t= linspace(0,this.psim.value,size(q,1))';
               else
                   nPts= size(q,1);
                   t= zeros(nPts, 1);
                   qpose= this.Pose(q);
                   qpose= qpose(:,1:3);
                   for i= 2:nPts
                      d= sqrt(sum((qpose(i-1,:)- qpose(i,:)).^2));
                      t(i)= t(i-1)+ d/this.psim.value+ 0.01; 
                   end
               end
               %opt= simset('SrcWorkspace','Current');
               %q0= q(1,:); % Posición inicial del robot
               opt= [];
               assignin('base', 'q0', q(1,:));
               if size(pose,1)<3 % Para dibujar trayectoria en pantalla
                   %pose= repmat(qpose(end,:),[3,1]).*(1+ randn(3,3)*1e-3);
                   pose= repmat(qpose(end,:),[3,1])+ randn(3,3)*1e-3;
               else
                   pose= pose(:,1:3);
               end
               assignin('base', 'pose', pose);
               sim(this.psim.mdl, t, opt, [t,q]);
               % Guarda valores para reproducción final
               if this.save.value==1
                  this.save.q= [this.save.q; q];
                  if isempty(this.save.t)
                     this.save.t= [this.save.t; t];
                  else
                     this.save.t= [this.save.t; t+ this.save.t(end)+1];
                  end
                  this.save.pose= [this.save.pose; pose];
               end
           end
           this.q= q(end,:);
        end
           
        function Sim(this, mdl, value, type)
            if nargin==1
                this.psim.mdl=[];
                return
            elseif nargin<3
                error('Sim(this, mdl, par, [type])');
            elseif nargin==3
                this.psim.type= 'V';
            else % 'T'
                this.psim.type= type;
            end
            this.psim.value= value;
            this.psim.mdl= mdl;   
        end
        
        function Rec(this, value) 
        % Control de grabación de simulación
            if nargin==1 % Resetea e inicia la grabación
                this.save.q= [];
                this.save.t= [];
                this.save.pose= [];
                this.save.value= 1;                
            elseif value==0 % Para la grabación
                this.save.value= 0;
            elseif value==1 % Inicia la grabación sin resetear
                this.save.value= 1;
            else
                error('Valor incorrecto');
            end
        end
        
        function Rep(this)
           if size(this.save.q,1)>3
               opt= simset('SrcWorkspace','Current');
               q0= this.save.q(1,:); % Posición inicial del memoria
               % Para evitar dos valores iguales.
               pose= this.save.pose+ randn(size(this.save.pose,1),3)*1e-5;
               sim(this.psim.mdl,this.save.t,opt,[this.save.t,this.save.q]);
           else
               disp('No hay record')
           end
        end
        
    end % methods
    
end % class